/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Defines the warehouse::Collection class
 *
 * \author Bhaskara Marthi
 */

#ifndef WAREHOUSE_COLLECTION_H
#define WAREHOUSE_COLLECTION_H

#include <warehouse/UpdateNotification.h>
#include <warehouse/result_iterator.h>
#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <boost/thread.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/scoped_ptr.hpp>
#include <string>

namespace warehouse
{

/// \brief Class used to query and update a collection in the warehouse
/// External code shouldn't directly create objects of this class: instead, use
/// WarehouseClient::setupCollection.
template <class M>
class Collection
{
public:

  /// \brief Constructor.  Should not be called by external code
  Collection (const std::string& topic, const std::string& db,
              const std::string& collection, const ros::NodeHandle& nh);

  /// \brief Copy constructor.  Should not be called by external code.
  /// Should only be used when rhs is about to go out of scope.
  Collection (const Collection& pub);

  /// \brief Default constructor.  Can be called by external code
  /// if you want to create a Collection object and assign it later
  Collection ();

  /// \brief Assignment operator
  Collection<M>& operator= (const Collection<M>& c);


  /// \brief Publishes a message with given metadata.  Waits for notification
  /// and returns the id of the message.
  /// \param metadata A yaml encoded dictionary of key-value pairs, where the keys are
  /// strings and the values are numbers or strings (use the yaml_cpp library to generate)
  /// \tparam M must equal the advertised type of this Collection
  std::string publish (const M& msg, const std::string& metadata = std::string());

  /// \brief Query the collection
  /// \tparam Message type.  Must match message type in \a collection
  /// \param metadata_only If true, only get the metadata, not the message objects
  /// themselves (i.e., default constructed message objects will be returned, which
  /// saves space if the actual messages include large arrays)
  /// \throws WarehouseClientException if any of the service calls involved fails
  /// \retval The return value is a pair of iterators, that models
  /// <a href="http://www.boost.org/doc/libs/1_46_0/libs/range/doc/html/range/concepts/single_pass_range.html">
  /// Boost single pass range</a>.  In other words, you can use the iterators in loops and STL operations,
  /// or use the returned range directly in something like BOOST_FOREACH.
  /// The iterators lazily pull messages from the warehouse as needed.
  ///
  /// Can conceivably get stuck if the warehouse dies during the interaction
  typename QueryResults<M>::range_t
  queryResults (const std::vector<Condition>& conditions,
                bool metadata_only = false,
                const OrderingCriterion& order = OrderingCriterion()) const;

  /// \brief Front-end to queryResults that fetches all the results
  /// and puts them into a vector.
  std::vector<typename MessageWithMetadata<M>::ConstPtr >
  pullAllResults (const std::vector<Condition>& conditions,
                  bool metadata_only = false,
                  const OrderingCriterion& order = OrderingCriterion()) const;

  /// \brief Remove messages from a collection
  /// \retval Number of messages removed
  /// \throw WarehouseClientException
  unsigned removeMessages (const std::vector<Condition>& conditions);

  /// \brief Get the collection name
  std::string collectionName () const;

private:

  /// Assumes lock is held on mutex_
  void addMetadata (const std::string& id, const std::string& metadata);

  // We'll wrap the state in a struct so that we can copy and assign Collections
  // (with move semantics)
  struct Impl : public boost::noncopyable
  {
    Impl (const string& topic, const string& db,
          const string& coll, const ros::NodeHandle& nh);

    void connectCallback (const ros::SingleSubscriberPublisher& pub);
    void notificationCallback (const UpdateNotification& update);
    void setupConnections ();


    boost::mutex mutex_;
    boost::condition_variable cond_;
    ros::NodeHandle nh_;
    boost::scoped_ptr<ros::CallbackQueue> cb_queue_;
    ros::AsyncSpinner spinner_;

    ros::Publisher pub_;
    ros::Subscriber sub_;

    // These are mutable so const methods can call the services
    mutable ros::ServiceClient query_client_, pull_client_, metadata_client_,
      remove_messages_client_;
    
    std::string topic_, db_, collection_;
    bool initialized_;

    std::string node_name_;

    // Used in setupCollection to tell that publisher has connected to warehouse
    bool warehouse_sub_connection_received_;
    boost::optional<std::string> last_id_;
  };

  typedef boost::shared_ptr<Impl> ImplPtr;
  ImplPtr impl_;
};



} // namespace


#include "../../src/collection_impl.h"

#endif // include guard
