/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */



#ifndef WAREHOUSE_EXCEPTION_H
#define WAREHOUSE_EXCEPTION_H

#include <boost/format.hpp>
#include <stdexcept>
#include <string>

namespace warehouse
{

using boost::format;
using std::string;

/// A base class for all warehouse client exceptions; provides a handy boost::format parent constructor
class WarehouseClientException: public std::logic_error
{
public:
  WarehouseClientException (const format& error_string) : std::logic_error(error_string.str()) {};
  WarehouseClientException (const char* str) : std::logic_error(str) {};
};

/// \brief Setup service error
struct SetupCollectionException: public WarehouseClientException
{
  SetupCollectionException (const string& msg) :
    WarehouseClientException (format ("Error setting up collection: %1%") % msg) {}
};

/// \brief Error adding metadata to a message
struct MetadataException: public WarehouseClientException
{
  MetadataException (const unsigned code, const string& msg) :
    WarehouseClientException (format ("Error %1% adding metadata: %2%") % code % msg) {}
};

/// \brief Key not found in metadata
struct MissingKeyException: public WarehouseClientException
{
  MissingKeyException (const string& metadata, const string& key) :
    WarehouseClientException (format ("Couldn't find key %1% in metadata %2%") % key % metadata) {}
};      


} // namespace

#endif // include guard
