/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Defines a class for a message together with string metadata, and
 * some utility functions to create and parse metadata objects
 *
 * \author Bhaskara Marthi
 */

#include <warehouse/exception.h>
#include <std_msgs/String.h>
#include <yaml-cpp/yaml.h>


#ifndef WAREHOUSE_MESSAGE_WITH_METADATA_H
#define WAREHOUSE_MESSAGE_WITH_METADATA_H

namespace warehouse
{

/// \brief Class that wraps (via inheritance) a ROS message type, together with
/// additional metadata (a yaml dictionary)
/// \tparam M the message being wrapped
template <class M>
struct MessageWithMetadata : public M
{
  MessageWithMetadata (const M& msg, const std::string& metadata) :
    metadata(metadata)
  {
    M::operator=(msg);
  }

  MessageWithMetadata (const MessageWithMetadata& m) :
    M(m), metadata(m.metadata)
  {}

  MessageWithMetadata ()
  {}

  std::string metadata;

  typedef boost::shared_ptr<MessageWithMetadata<M> > Ptr;
  typedef boost::shared_ptr<MessageWithMetadata<M> const> ConstPtr;
};



/// \brief Class that simplifies the creation of yaml strings representing
/// metadata
///
/// Usage:
/// std::string meta = MetadataString().add("foo", "foo-val").add("bar", 42);
class MetadataString
{
public:
  
  MetadataString ()
  {
    done = false;
    out << YAML::BeginMap;
  }

  template <typename T>
  MetadataString& add (const std::string& key,
                 const T& value)
  {
    ROS_ASSERT_MSG (!done, "Can't use this metadata object any more");
    out << YAML::Key << key;
    out << YAML::Value << value;
    return *this;
  }

  operator std::string ()
  {
    out << YAML::EndMap;
    done = true;
    return std::string(out.c_str());
  }
  

private:
  
  YAML::Emitter out;
  bool done;
};


template <typename T>
T getMetadataVal (const std::string& metadata, const std::string& key)
{
  std::stringstream s(metadata, std::ios_base::in);
  YAML::Parser parser(s);
  YAML::Node doc;
  const bool success = parser.GetNextDocument(doc);
  ROS_ASSERT_MSG(success, "Error parsing metadata %s", metadata.c_str());
  if (const YAML::Node* p_name = doc.FindValue(key))
  {
    T val;
    *p_name >> val;
    return val;
  }
  else
  {
    throw MissingKeyException(metadata, key);
  }
}


} // namespace

#endif // include guard
