/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Results of a query
 *
 * \author Bhaskara Marthi
 */

#ifndef WAREHOUSE_RESULT_SET_H
#define WAREHOUSE_RESULT_SET_H

#include <warehouse/message_with_metadata.h>
#include <warehouse/InitiateQuery.h>
#include <boost/iterator/iterator_facade.hpp>
#include <boost/utility.hpp>
#include <boost/thread.hpp>
#include <ros/ros.h>


namespace warehouse
{

/// \brief An iterator over results of a query
///
/// This type models
/// <a href="http://www.boost.org/doc/libs/1_46_0/libs/iterator/doc/new-iter-concepts.html">
/// Boost single pass iterator
/// </a>, over MessageWithMetadata<M> objects.
///
/// Note: none of the operations on this iterator are threadsafe due to the
/// underlying ros calls
template <class M>
class ResultIterator :
    public boost::iterator_facade<ResultIterator<M>,
                                  typename MessageWithMetadata<M>::ConstPtr,
                                  boost::single_pass_traversal_tag,
                                  typename MessageWithMetadata<M>::ConstPtr >
{
public:

  // Handy typedef
  typedef MessageWithMetadata<M> Msg;
  typedef typename M::ConstPtr Ptr;

  /// \brief Constructor; not to be called by outside code
  ResultIterator (const ros::NodeHandle nh, const InitiateQuery& srv,
                  ros::ServiceClient& pull_client, bool metadata_only);

  /// \brief Copy constructor; not to be called by outside code
  /// Only call when rhs is about to go out of scope
  ResultIterator (const ResultIterator& res_);

  /// \brief Constructor for past_the_end iterator.
  /// Not to be called by outside code.
  ResultIterator ();

private:

  friend class boost::iterator_core_access;

  /// Member functions needed to be an iterator
  void increment ();
  typename MessageWithMetadata<M>::ConstPtr dereference () const;
  bool equal (const ResultIterator<M>& other) const;
  void maybeFetchFirst () const;

  struct QueryResultSubscriber : boost::noncopyable
  {
    QueryResultSubscriber (const ros::NodeHandle& nh, const InitiateQuery& srv,
                           const ros::ServiceClient& pull_client_,
                           bool metadata_only);
    
    void messageCallback (Ptr msg);
    void metadataCallback (std_msgs::String::ConstPtr m);
    void fetchNext ();

    boost::mutex mutex_;
    boost::condition_variable cond;
    ros::NodeHandle nh_;
    std::string query_id_, topic_base_;
    bool data_received_;
    bool metadata_received_;
    ros::Subscriber sub_, metadata_sub_;
    ros::ServiceClient pull_client_;
    bool metadata_only_;
    bool initial_fetch_;
    typename MessageWithMetadata<M>::Ptr next_result_;
  };

  mutable boost::shared_ptr<QueryResultSubscriber> sub_;
};


/// A templated typedef for convenience
template <class M>
struct QueryResults
{
  typedef std::pair<ResultIterator<M>, ResultIterator<M> > range_t;
};


} // namespace

#include "../../src/result_iterator_impl.h"

#endif // include guard
