/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * C++ client library that wraps the Warehouse ROS API
 *
 * \author Bhaskara Marthi
 */


#include <warehouse/collection.h>
#include <warehouse/message_with_metadata.h>
#include <warehouse/result_iterator.h>
#include <warehouse/UpdateNotification.h>
#include <ros/ros.h>
#include <string>
#include <boost/thread.hpp>
#include <boost/optional.hpp>
#include <boost/utility.hpp>

#ifndef WAREHOUSE_WAREHOUSE_CLIENT_H
#define WAREHOUSE_WAREHOUSE_CLIENT_H

namespace warehouse
{

typedef std::vector<std::string> StringVec;


/// \brief Class that wraps the warehouse ROS API.
/// 
/// Threadsafe.
/// Does it's own spinning and callback queue management, so
/// external code doesn't need to add any additional spinner threads.
class WarehouseClient : boost::noncopyable
{
public:
  
  /// \brief Empty constructor
  /// \param db The name of the database in the warehouse that this client will
  /// talk to
  WarehouseClient (const string& db);

  /// \brief Connect to a collection in the warehouse.  Create the collection if
  /// \brief Connect to a collection in the warehouse.  Create the collection if
  /// it doesn't exist, otherwise, check that the type and indexed fields match.
  /// This version uses an existing ROS topic as the source for new messages.
  /// \tparam M Type of message
  /// \param topic Topic on which collection will listen for new messages
  /// \throws WarehouseClientException if the service call fails at the ROS
  /// level (e.g., the warehouse node doesn't exist)
  /// \throws SetupCollectionException if the service call returns with an error
  template <class M>
  Collection<M> setupCollection (const std::string& collection,
                                 const std::string& topic, 
                                 const StringVec& indexed_fields = StringVec(),
                                 const StringVec& additional_metadata_fields =
                                 StringVec());


  /// it doesn't exist, otherwise, check that the type and indexed fields match.
  /// \tparam M Type of message
  /// \throws WarehouseClientException if the service call fails at the ROS
  /// level (e.g., the warehouse node doesn't exist)
  /// \throws SetupCollectionException if the service call returns with an error
  template <class M>
  Collection<M> setupCollection (const std::string& collection,
                                 const StringVec& indexed_fields = StringVec(),
                                 const StringVec& additional_metadata_fields =
                                 StringVec());


  /// \brief Get list of collections currently in db
  std::vector<std::string> listCollections () const;

  /// \retval Name of the db this client talks to
  std::string dbName () const;
    
private:

  mutable boost::mutex mutex_; // Mutable since const operations can acquire it
  const std::string db_;
  ros::NodeHandle nh_;
  ros::ServiceClient setup_client_;
  mutable ros::ServiceClient list_collections_client_;

};

} // namespace

#include "../../src/warehouse_client_impl.h"

#endif // include guard

