#!/usr/bin/env python
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import roslib; roslib.load_manifest('warehouse')
import pymongo as pm
from pymongo import Connection
from pymongo import json_util
from pymongo import objectid
import bson
import gridfs
import rospy
import warehouse.srv as whs
import warehouse.msg as whm
import StringIO
import threading
import roslib.rostime as rostime
import std_msgs.msg
import json
import yaml
import gc

from rospy import loginfo

DELIMETER='___'

############################################################
# Utilities
############################################################


def legal_db_name(name):
    return not name in ['admin', 'local'] and \
            name and not name.startswith('_')


def legal_collection_name(name):
    return not name in ['system.indexes', 'fs.chunks'] and \
            name and not name.startswith('_')

def legal_topic_name(name):
    return name.startswith('/')

def updatable_metadata_field(name):
    return name and not name.startswith('_')

def extract_nested_field(msg, f):
    """
    Helper function that extracts a nested field from a ros message
    """
    for field in f:
        msg = getattr(msg, field)
    return msg
  
def parse_val(s, example):
    """
    Parse a string val based on the type of example
    """
    if isinstance(example, str) or isinstance(example, unicode):
        return s
    elif isinstance(example, int):
        return int(s)
    elif isinstance(example, float):
        try:
            return float(s)
        except ValueError:
            return s
    else:
        print ("Example {0} has type {1}".format(example, type(example)))
        raise ValueError(example)  
    
def construct_query(q, example):
    """
    Helper to construct a mongo db query
    """
    query = {}
    for c in q:
        fields = c.field_name.split('.')
        db_field = DELIMETER.join(fields)

        # default to float.  This happens in the case
        # where the field is part of the metadata, and due
        # to a race condition, it isn't present in the example.
        dummy_val = example.get(db_field, 0.0)

        if c.predicate == whm.Condition.EQUALS:
            v = parse_val(c.args[0], dummy_val)
        elif c.predicate == whm.Condition.LESS_THAN:
            v = { "$lt" : parse_val(c.args[0], dummy_val) } 
        elif c.predicate == whm.Condition.GREATER_THAN:
            v = { "$gt" : parse_val(c.args[0], dummy_val) }
        else:
            raise ValueError(c)
        query[db_field] = v
    return query

def doc_to_string(d):
    return json.dumps(d, default=json_util.default)

def metadata_resp(code, msg):
    rospy.logerr("Handle metadata returned error {0}: {1}".\
                 format(code, msg))
    return whs.MetadataResponse(error_code=code, error_msg=msg)


############################################################
# Handler for individual collection
############################################################

class CollectionHandler:
    """
    A class that handles insertion of data in a collection.
    """

    def __init__(self, req, msg_class, conn, topic):
        """
        Constructor. Listens to a topic named
        /warehouse/<db_name>/<collection_name>/insert.
        Inserts data of the type <msg_type> when it comes in over the wire.
        
        @param req: The SetupCollection service request
        @param msg_type: The message data type
        @param conn: The connection to the db
        @param topic: The topic to push data on

        @todo Should check the indexed fields for correctness here rather than
        dying in the callback
        """
        # Basic setup
        self.connection = conn 
        self.db = conn[req.db_name]
        self.collection = self.db[req.collection_name]
        self.fs = gridfs.GridFS(self.db)
        self.cursors = {}
        self.query_pubs = {}
        self.query_metadata_pubs = {}
        self.msg_class = msg_class
        self.publish_messages = {}

        # Create a topic in which to push data
        notification_topic = '~{0}/{1}/notify'.format(req.db_name,
                                                      req.collection_name)
        self.indexed_fields = [f.split('.') for f in req.indexed_fields if \
                               f not in req.additional_metadata_fields]
        
        self.sub = rospy.Subscriber(topic, rospy.AnyMsg, self.handle_push)

        # Topic to send notifications upon new data
        self.pub = rospy.Publisher(notification_topic,
                                   whm.UpdateNotification)

    def cleanup(self):
        self.sub.unregister()
        self.pub.unregister()

    def handle_push(self, raw_msg):
        """
        Handles the insertion of data into a collection.
        """

        # Store the raw message in the DB's GridFS instance 
        msg_id = self.fs.put(raw_msg._buff)

        # Create db entry
        db_entry = {}
        db_entry['_blob_id'] = msg_id

        # Extract indexed fields
        if self.indexed_fields:
            msg = self.msg_class()
            msg.deserialize(raw_msg._buff)
        
            for f in self.indexed_fields:
                val = extract_nested_field(msg, f)
                if isinstance(val, rostime.Time) or \
                       isinstance(val, rostime.Duration):
                    # Times or durations are converted to a floating point
                    # number measured in seconds before storing in metadata
                    val = val.to_sec()
                db_entry[DELIMETER.join(f)] = val

        # Store the creation time
        db_entry['_creation_time'] = rospy.Time.now().to_sec()

        # Store the indexing information in the table
        self.collection.insert(db_entry)

        # Send out a notification
        self.pub.publish(id=str(db_entry['_id']),
                         update_type=whm.UpdateNotification.NEW_MESSAGE,
                         sender=raw_msg._connection_header['callerid'], 
                         metadata=doc_to_string(db_entry))

        

    def initiate_query(self, req, id):
        # We use a prototype from the collection to infer types and
        # figure out how to parse the fields of the query
        example = self.collection.find_one()
        
        # Special case: if there's no prototype, the collection is empty
        # Which is lucky, because construct_query would have failed

        if example is None:
            self.cursors[id] = self.collection.find()
        
        else:
            self.cursors[id] = self.collection.find(
                spec=construct_query(req.query, example),
                sort=construct_order(req.order))
            
        self.query_pubs[id] = rospy.Publisher(
            '~{0}/{1}/{2}'.format(req.db_name, req.collection_name, id),
            self.msg_class, latch=True)
        self.query_metadata_pubs[id] = rospy.Publisher(
            '~{0}/{1}/metadata/{2}'.format(req.db_name, req.collection_name, id),
            std_msgs.msg.String, latch=True)
        self.publish_messages[id] = not req.metadata_only

        return whs.InitiateQueryResponse(query_id=id)

    def pull_message(self, id):
        try:
            item = self.cursors[id].next()
            m = rospy.AnyMsg()
            m._buff = self.fs.get(item['_blob_id']).read()
            if self.publish_messages[id]:
                self.query_pubs[id].publish(m)
            self.query_metadata_pubs[id].publish(doc_to_string(item))
            return 1
        except StopIteration:
            self.query_pubs[id].unregister()
            self.query_metadata_pubs[id].unregister()
            del self.cursors[id]
            del self.publish_messages[id]
            del self.query_pubs[id]
            del self.query_metadata_pubs[id]
            return 0

    def handle_remove(self, cond):
        # Use a prototype to parse query as in initiate_query
        example = self.collection.find_one()
        if example is None:
            return 0

        # Now guaranteed that example is valid
        q = construct_query(cond, example)

        # First identify the underlying ros message blobs
        # We do it this way to never be in a state where a doc is in the
        # db but the underlying message has been deleted.
        # But it does require storing all the blob ids in memory at once
        cursor = self.collection.find(q)
        blobs = [item['_blob_id'] for item in cursor]

        # Remove the collection items
        self.collection.remove(q)

        # Remove the messages
        for b in blobs:
            self.fs.delete(b)

        return len(blobs)
        

    def handle_metadata(self, req):
        for i in range(50):
            item = self.collection.find_one(objectid.ObjectId(req.id))
            if item is not None:
                if i>0:
                    rospy.logerr("Found {0} in {1}/{2} on attempt {3}".\
                                 format(req.id, req.db_name,
                                        req.collection_name, i))
                break
            rospy.logerr("Couldn't find {0} in {1}/{2} (attempt {3})".\
                         format(req.id, req.db_name, req.collection_name, i))
            rospy.sleep(0.01)
        if item is None:
            return metadata_resp(whs.MetadataResponse.INVALID_ID,
                                 "Couldn't find item {0} in {1}/{2}".\
                                 format(req.id, req.db_name, 
                                        req.collection_name))

        # Parse the json
        try:
            meta = yaml.load(req.metadata)
        except ValueError as e:
            return metadata_resp(whs.MetadataResponse.JSON_ERROR,
                                 "Error '{0}' parsing metadata: {1}".\
                                 format(e, req.metadata))

        # Update the item and save
        for k, v in meta.items():
            if updatable_metadata_field(k):
                item[k] = v
            else:
                rospy.warn("Cannot modify value of metadata field {0}".\
                           format(k))
        self.collection.save(item)

        # Publish notification and return
        self.pub.publish(id=req.id,
                         update_type=whm.UpdateNotification.UPDATE_METADATA,
                         sender = req._connection_header['callerid'],
                         metadata=doc_to_string(item))
        return whs.MetadataResponse(error_code=whs.MetadataResponse.SUCCESS)
                                    
        

def construct_order(order):
    """
    Construct a Mongodb ordering criterion from Ros message.
    If the field is empty, return None.
    """
    if order.field:
        if order.field == '_creation_time':
            db_field = '_creation_time'
        else:
            fields = order.field.split('.')
            db_field = DELIMETER.join(fields)
        return [(db_field, pm.DESCENDING if order.reverse else pm.ASCENDING)]


############################################################
# The main warehouse backend
############################################################

class Backend:
    """
    The mongo database interface. This serves as the backend to the warehouse.
    """
    def __init__(self, host, port):
        """
        Constructor for the DB Interface. This creates a connection to the MonogDB
        and establishes a set of services
        """
        done = False
        self.handlers = {}

        self.host = host
        self.port = port

        # Create a connection to the mongo DB
        while not done and not rospy.is_shutdown():
            done = True
            try:
                loginfo( "Attempting to connect to mongodb @ {0}:{1}".\
                               format(self.host,self.port))
                self.connection = Connection(self.host, self.port)
            except:
                done = False
                rospy.sleep(2.0)

        if rospy.is_shutdown():
            rospy.logwarn( "Warehouse backend has been killed" )
            return

        loginfo( "Connected to mongodb" )

        self.lock = threading.Lock()
        self.next_query_id = 0
        self.query_handlers = {}

        self.close_collection_srv = rospy.Service('close_collection',
                                                  whs.CloseCollection,
                                                  self.handle_close_collection)

        # Create a service that checks for the existence of a collection
        self.list_collections_srv = rospy.Service('list_collections',
                                                  whs.ListCollections,
                                                  self.list_collections)

        # Create a service to handle collection creation
        self.setup_collection = rospy.Service('setup_collection',
                                              whs.SetupCollection,
                                              self.handle_setup_collection)

        # Service for initiating a query
        self.initiate_query_srv = rospy.Service('initiate_query',
                                                whs.InitiateQuery,
                                                self.handle_query)

        # Service for pulling messages
        self.pull_message_srv = rospy.Service('pull_message',
                                              whs.PullMessage,
                                              self.handle_pull)

        # Service for dropping a db
        self.drop_db = rospy.Service('drop_db',
                                     whs.DropDB,
                                     self.handle_drop)

        # Service for metadata
        self.metadata_srv = rospy.Service('add_metadata',
                                          whs.Metadata,
                                          self.handle_metadata)

        # Service for removing messages
        self.remove_msg_srv = rospy.Service('remove_messages',
                                            whs.RemoveMessages,
                                            self.handle_remove)

    def handle_remove(self, req):
        def resp(code, msg, num=0):
            return whs.RemoveMessagesResponse(error_code=code, error_msg=msg,
                                              num_messages_removed=num)

        with self.lock:
            try:
                db = self.handlers[req.db_name]
            except KeyError:
                return resp(whs.RemoveMessagesResponse.INVALID_DB_NAME,
                            "Db {0} not found".format(req.db_name))
            try:
                h = db[req.collection_name]
            except KeyError:
                return resp(whs.RemoveMessagesResponse.INVALID_COLLECTION_NAME,
                            "Collection {0} not found".format(req.collection_name))
            response = resp(whs.RemoveMessagesResponse.SUCCESS, '',
                            h.handle_remove(req.conditions))
            return response
                                                   
        

    def handle_metadata(self, req):
        with self.lock:
            try:
                db = self.handlers[req.db_name]
            except KeyError:
                return metadata_resp(whs.MetadataResponse.INVALID_DB_NAME,
                                     "Db {0} not found".format(req.db_name))

            try:
                h = db[req.collection_name]
            except KeyError:
                return metadata_resp(whs.MetadataResponse.INVALID_COLLECTION_NAME,
                                     "Collection {0} not present in {1}".\
                                     format(req.collection_name, req.db_name))

            resp = h.handle_metadata(req)
            return resp

        

    def handle_drop(self, req):
        """
        Drop a database
        """
        with self.lock:
            assert(not req.db_name in self.handlers)
            self.connection.drop_database(req.db_name)
            loginfo('dropped %s', req.db_name)
            return whs.DropDBResponse() 

    def handle_pull(self, req):
        """
        Cause next message on a query to be published
        """
        with self.lock:
            n = self.query_handlers[req.query_id].pull_message(req.query_id)
            if n == 0:
                del self.query_handlers[req.query_id]
            return whs.PullMessageResponse(num_messages_sent=n)


    def handle_query(self, req):
        """
        Pass off the query to the appropriate collection
        """
        with self.lock:
            query_id = 'messages_{0}'.format(self.next_query_id)
            self.next_query_id += 1
            handler = self.handlers[req.db_name][req.collection_name]
            self.query_handlers[query_id] = handler
            resp = handler.initiate_query(req, query_id)
            return resp


    def get_db(self, name):
        """
        Create db if necessary.
        """
        created = False
        if name not in self.connection.database_names():
            # need to do this because of Mongo's laziness
            self.connection[name].create_collection('___DUMMY_TABLE___') 
            loginfo('Created database %s', name)
            created = True

        # Because Mongo documentation is not completely clear on this
        assert name in self.connection.database_names()
        resp = (self.connection[name], created)
        return resp

    def handle_close_collection(self, req):
        with self.lock:
            d_name = req.db_name
            c_name = req.collection_name
            if d_name in self.handlers and c_name in self.handlers[d_name]:
                self.handlers[d_name][c_name].cleanup()
                del self.handlers[d_name][c_name]
            if len(self.handlers[d_name]) == 0:
                del self.handlers[d_name]
            return whs.CloseCollectionResponse()

    def list_collections(self, req):
        """
        Service callback that handles db and collection existence checks
        @param req: The whs.ListCollections service request
        """
        with self.lock:
            db_exists = False
            collections = []

            if req.db_name in self.connection.database_names():
                db_exists = True
                collections = [c['name'].encode() for c in \
                               self.connection[req.db_name]._ros_collection_info.find()]

            return whs.ListCollectionsResponse(db_exists=db_exists,
                                               collections=collections)

    def handle_setup_collection(self, req):
        """
        Service callback that handles setting up new collection
        @param req: The whs.SetupCollection service request
        """
        from warehouse.srv import SetupCollectionResponse

        with self.lock:
            def error_msg(ec, msg):
                return SetupCollectionResponse(error_code=ec, error_msg=msg)

            # Resolve topic
            topic = req.topic or rospy.resolve_name\
                    ('~' + req.db_name +'/'+ req.collection_name + '/insert')

            # Check the names
            if not legal_db_name(req.db_name):
                return error_msg(SetupCollectionResponse.ILLEGAL_DB_NAME,
                                 'Illegal db name {0}'.format(req.db_name))
            if not legal_collection_name(req.collection_name):
                return error_msg(SetupCollectionResponse.ILLEGAL_COLLECTION_NAME,
                                 'Illegal collection name {0}'.\
                                 format(req.collection_name))
            if not legal_topic_name(topic):
                return error_msg(SetupCollectionResponse.ILLEGAL_TOPIC_NAME,
                                 '{0} was not a globally qualified topic name'.\
                                 format(topic))


            try:
                # Import the module
                roslib.load_manifest(req.msg_pkg)
                m = __import__(req.msg_pkg+".msg")
                mod = getattr(m, 'msg')
            except roslib.packages.InvalidROSPkgException:
                return error_msg(whs.SetupCollectionResponse.PACKAGE_NOT_FOUND, \
                                 'package {0} not found'.format(req.msg_pkg))
            except ImportError:
                return error_msg(SetupCollectionResponse.MESSAGE_NOT_FOUND_IN_PACKAGE,
                                 'no messages found in {0}'.format(req.msg_pkg))
            rospy.logdebug('Imported %s', mod)

            # Look up message class
            try:
                cls = getattr(mod, req.msg_type)
            except AttributeError:
                return error_msg(whs.SetupCollectionResponse.MESSAGE_NOT_FOUND_IN_PACKAGE, \
                                 'message type {0} not found in {1}'.format(req.msg_type, req.msg_package))
            loginfo('Looked up class %s', mod)

            # If the db and collection exist, verify that relevant info matches.  Else create them.
            db, db_created = self.get_db(req.db_name)
            c = req.collection_name
            collection_created = False
            info = db._ros_collection_info.find_one({'name' : c})
            dummy = cls()
            if info: 
                existing_info = (info['msg_pkg'], info['msg_type'], info['indexed_fields'], info['md5'])
                req_info = (req.msg_pkg, req.msg_type, req.indexed_fields, dummy._md5sum)
                if existing_info != req_info:
                    return error_msg(whs.SetupCollectionResponse.EXISTING_COLLECTION_MISMATCH, 
                                     'Args {0} did not match saved info {1}'.format(req_info, existing_info))
            else:
                # Check the index fields 
                for f in req.indexed_fields:
                    if f not in req.additional_metadata_fields:
                        try:
                            extract_nested_field(dummy, f.split('.'))
                        except AttributeError:
                            return error_msg(whs.SetupCollectionResponse.INVALID_INDEXED_FIELD,
                                             msg='Field {0} not found in message'.format(f))

                # Create an entry in the _ros_collection_info table
                doc = {'name': c, 'msg_pkg': req.msg_pkg, 'msg_type': req.msg_type, \
                       'indexed_fields': req.indexed_fields, 'md5': dummy._md5sum, 'fulltext': dummy._full_text}
                db._ros_collection_info.insert(doc)
                collection_created = True
                loginfo('Created collection {3}/{2} for message type {0}/{1}'.\
                              format(doc['msg_pkg'], doc['msg_type'], c, req.db_name))

                # Create indices
                for f in req.indexed_fields:
                    name = DELIMETER.join(f.split('.'))
                    db[c].create_index(name, name=name)
                    loginfo('Generated index on {0}'.format(name))

            # Create an object to handle data push events if one doesn't already exist
            if not req.db_name in self.handlers:
                self.handlers[req.db_name] = {}

            if not c in self.handlers[req.db_name]:
                self.handlers[req.db_name][c] = CollectionHandler(req, cls, self.connection, topic) 
                loginfo('Created handler')


            return whs.SetupCollectionResponse(\
                error_code=whs.SetupCollectionResponse.SUCCESS, db_created=db_created,
                collection_created=collection_created, topic=topic)

if __name__ == '__main__':
    rospy.init_node('warehouse')
    host = rospy.get_param('~host', 'localhost')
    port = rospy.get_param('~port', 27017)
    back = Backend(host, port)
    rospy.spin()
