/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Implementation of template methods for Collection
 * Should only be included by collection.h
 *
 * \author Bhaskara Marthi
 */

#include <warehouse/result_iterator.h>
#include <warehouse/Metadata.h>
#include <warehouse/PullMessage.h>
#include <warehouse/InitiateQuery.h>
#include <warehouse/RemoveMessages.h>
#include <yaml-cpp/yaml.h>

namespace warehouse
{

using std::string;

/************************************************************
 * Initialization
 ***********************************************************/


template <class M>
Collection<M>::Collection (const std::string& topic, const string& db,
                           const string& coll, const ros::NodeHandle& nh) :
  impl_(new Collection<M>::Impl(topic, db, coll, nh))
{
}

template <class M>
Collection<M>::Collection (const Collection<M>& c) :
  impl_(c.impl_)
{
}

template <class M>
Collection<M>::Collection ()
{}

template <class M>
Collection<M>&
Collection<M>::operator= (const Collection<M>& c)
{
  impl_ = c.impl_;
  return *this;
}

template <class M>
Collection<M>::Impl::Impl (const string& topic, const string& db,
                           const string& coll, const ros::NodeHandle& nh) :
  nh_(nh), cb_queue_(new ros::CallbackQueue()), spinner_(1, cb_queue_.get()),
  query_client_(nh_.serviceClient<InitiateQuery>("initiate_query")),
  pull_client_(nh_.serviceClient<PullMessage>("pull_message")),
  metadata_client_(nh_.serviceClient<Metadata>("add_metadata")),
  remove_messages_client_(nh_.serviceClient<RemoveMessages>("remove_messages")),
  topic_(topic), db_(db), collection_(coll), initialized_(false),
  node_name_(ros::this_node::getName()),
  warehouse_sub_connection_received_(false)
{
  nh_.setCallbackQueue(cb_queue_.get());
  spinner_.start();
  setupConnections();
}

inline
void waitForService (ros::ServiceClient& client)
{
  while (ros::ok()) {
    if (client.waitForExistence(ros::Duration(5.0)))
      return;
    ROS_INFO_NAMED("init", "Waiting for service %s to exist",
                   client.getService().c_str());
  }
}





template <class M>
void Collection<M>::Impl::setupConnections ()
{
  ROS_ASSERT (!initialized_);
  pub_ = nh_.advertise<M>(topic_, 10, boost::bind(&Collection<M>::Impl::\
                                                  connectCallback, this, _1));

  // Wait for connection from warehouse
  ros::Rate r(10);
  while (!warehouse_sub_connection_received_ && ros::ok()) {
    ROS_DEBUG_STREAM_NAMED ("setup_collection", "Waiting for warehouse connection");
    r.sleep();
    ros::spinOnce();
  }

  const string notification_topic = "warehouse/" + db_ + "/" + collection_ + "/notify";
  sub_ = nh_.subscribe(notification_topic, 10,
                       &Collection<M>::Impl::notificationCallback, this);

  // We don't have to be as careful about checking the identity of the publisher because
  // the topic is in the private namespace of the warehouse.  Only a fool would publish
  // to such a topic.
  while (sub_.getNumPublishers()==0) {
    ROS_DEBUG_STREAM_NAMED ("setup_collection",
                            "Waiting for connection to warehouse on " <<
                            notification_topic);
    r.sleep();
    ros::spinOnce();
  }

  ROS_DEBUG_STREAM_NAMED ("setup_collection", "Waiting for services to exist");
  waitForService(query_client_);
  waitForService(pull_client_);
  waitForService(metadata_client_);
  waitForService(remove_messages_client_);
  initialized_ = true;
  ROS_DEBUG_STREAM_NAMED ("setup_collection", "Collection has been set up");
}


/************************************************************
 * Callbacks from warehouse
 ***********************************************************/

template <class M>
void Collection<M>::Impl::notificationCallback (const UpdateNotification& update)
{
  ROS_DEBUG_NAMED ("publish_int", "In notification callback");
  boost::lock_guard<boost::mutex> l(mutex_);
  if (update.update_type == UpdateNotification::NEW_MESSAGE &&
      update.sender == node_name_) {
    ROS_DEBUG_STREAM_NAMED ("publish_int", "Received ack from " << node_name_);
    last_id_ = update.id;
    cond_.notify_all();
  }
  else
    ROS_DEBUG_STREAM_NAMED ("publish_int", "Ignoring update " << update <<
                            " in notification callback");
}


template <class M>
void Collection<M>::Impl::connectCallback (const ros::SingleSubscriberPublisher& pub)
{
  boost::mutex::scoped_lock l(mutex_);
  const string sub_name = pub.getSubscriberName();
  if (sub_name == nh_.resolveName("warehouse")) {
    warehouse_sub_connection_received_ = true;
    ROS_DEBUG_STREAM_NAMED ("connect_callback", "Received connection callback from " << sub_name);
  }
  else
    ROS_WARN_STREAM_NAMED ("connect_callback", "Received subscriber connection from " <<
                           sub_name << "; assuming this is not the actual warehouse");
}


/************************************************************
 * Querying
 ***********************************************************/

template <class M>
typename QueryResults<M>::range_t
Collection<M>::queryResults (const vector<Condition>& conditions,
                             const bool metadata_only,
                             const OrderingCriterion& order) const
{
  ROS_ASSERT(impl_ && impl_->initialized_);
  boost::mutex::scoped_lock l(impl_->mutex_);
  
  // Construct service request
  InitiateQuery srv;
  srv.request.db_name = impl_->db_;
  srv.request.collection_name = impl_->collection_;
  srv.request.query = conditions;
  srv.request.order = order;
  srv.request.metadata_only = metadata_only;

  // Call service
  if (!impl_->query_client_.call(srv))
    throw WarehouseClientException("Service call to initiate_query failed");
  ROS_DEBUG_STREAM_NAMED ("initiate_query", "Service call to initiate query succeeded");

  // Set up subscription for responses
  return typename QueryResults<M>::range_t
    (ResultIterator<M>(impl_->nh_, srv, impl_->pull_client_, metadata_only),
     ResultIterator<M>());
}

template <class M>
vector<typename MessageWithMetadata<M>::ConstPtr >
Collection<M>::pullAllResults (const vector<Condition>& conditions,
                               const bool metadata_only,
                               const OrderingCriterion& order) const
{
  ROS_ASSERT(impl_ && impl_->initialized_);
  typename QueryResults<M>::range_t res = queryResults(conditions, metadata_only, order);
  return vector<typename MessageWithMetadata<M>::ConstPtr > (res.first, res.second);
}



/************************************************************
 * Update
 ***********************************************************/

inline
string emptyMetadata ()
{
  YAML::Emitter out;
  out << YAML::BeginMap << YAML::EndMap;
  return string(out.c_str());
}



template <class M>
string Collection<M>::publish (const M& msg, const string& metadata)
{
  ROS_ASSERT(impl_ && impl_->initialized_);
  ROS_DEBUG_NAMED ("publish_int", "Acquiring unique lock");
  boost::unique_lock<boost::mutex> l(impl_->mutex_);
  impl_->last_id_.reset();
  ROS_DEBUG_NAMED ("publish_int", "Publishing message");
  impl_->pub_.publish(msg);
  while (!impl_->last_id_ && ros::ok()) {
    ROS_DEBUG_NAMED ("publish_int", "Waiting on condition variable");
    impl_->cond_.wait(l);
  }
  addMetadata(*impl_->last_id_, metadata.size() > 0 ? metadata : emptyMetadata());
  ROS_DEBUG_NAMED ("publish_int", "Added metadata");
  return *impl_->last_id_;
}

template <class M>
void Collection<M>::addMetadata (const string& id, const string& metadata)
{
  ROS_ASSERT(impl_ && impl_->initialized_);
  Metadata srv;
  srv.request.db_name = impl_->db_;
  srv.request.collection_name = impl_->collection_;
  srv.request.id = id;
  srv.request.metadata = metadata;

  // Call service
  if (!impl_->metadata_client_.call(srv))
    throw WarehouseClientException("Service call to add_metadata failed");

  // Pass on errors
  if (srv.response.error_code != Metadata::Response::SUCCESS)
    throw MetadataException(srv.response.error_code, srv.response.error_msg);
}


template <class M>
unsigned
Collection<M>::removeMessages (const vector<Condition>& conditions)
{
  ROS_ASSERT(impl_ && impl_->initialized_);
  boost::mutex::scoped_lock l(impl_->mutex_); 
  RemoveMessages srv;
  srv.request.db_name = impl_->db_;
  srv.request.collection_name = impl_->collection_;
  srv.request.conditions = conditions;

  if (!impl_->remove_messages_client_.call(srv))
    throw WarehouseClientException("Service call to remove_messages failed");
  if (srv.response.error_code != RemoveMessages::Response::SUCCESS) 
    throw WarehouseClientException
      (format("Call to remove_messages returned error %1%: '%2%'") %
        srv.response.error_code % srv.response.error_msg);
  return srv.response.num_messages_removed;
}

/************************************************************
 * Accessors
 ***********************************************************/

template <class M>
string Collection<M>::collectionName () const
{
  return impl_->collection_;
}



} // namespace
