/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Implementation of template methods of ResultIterator
 * Only to be included by result_iterator.h
 *
 * \author Bhaskara Marthi
 */

#include <warehouse/PullMessage.h>
#include <warehouse/exception.h>

namespace warehouse
{

using std::vector;
using std::string;

/// Main constructor
template <class M>
ResultIterator<M>::ResultIterator (const ros::NodeHandle nh, const InitiateQuery& srv,
                                   ros::ServiceClient& pull_client,
                                   const bool metadata_only) :
  sub_(new ResultIterator<M>::QueryResultSubscriber(nh, srv, pull_client, metadata_only))
{}

template <class M>
ResultIterator<M>::QueryResultSubscriber::QueryResultSubscriber (const ros::NodeHandle& nh,
                                                                 const InitiateQuery& srv,
                                                                 const ros::ServiceClient& pull_client,
                                                                 const bool metadata_only) :
  nh_(nh), query_id_(srv.response.query_id),
  topic_base_("warehouse/" + srv.request.db_name + "/" + srv.request.collection_name + "/"),
  data_received_(true),
  sub_(nh_.subscribe(topic_base_ + query_id_, 10,
                     &ResultIterator<M>::QueryResultSubscriber::messageCallback, this)),
  metadata_sub_(nh_.subscribe(topic_base_ + "metadata/" + query_id_, 10,
                              &ResultIterator<M>::QueryResultSubscriber::metadataCallback, this)),
  pull_client_(pull_client), metadata_only_(metadata_only), initial_fetch_(false)
{
}

/// Represents a past-the-end iterator
template <class M>
ResultIterator<M>::ResultIterator () 
{
}

/// Copy constructor
template <class M>
ResultIterator<M>::ResultIterator (const ResultIterator<M>& res) :
  sub_(res.sub_)

{
}

template <class M>
void
ResultIterator<M>::maybeFetchFirst () const
{
  ROS_ASSERT_MSG(sub_, "Code assumption violated");
  if (!sub_->initial_fetch_) {
    sub_->fetchNext();
    sub_->initial_fetch_ = true;
  }
}


template <class M>
typename MessageWithMetadata<M>::ConstPtr 
ResultIterator<M>::dereference()  const
{
  maybeFetchFirst();
  ROS_ASSERT_MSG(sub_->next_result_, "Can't dereference iterator: no more results");
  return sub_->next_result_;
}

template <class M>
void ResultIterator<M>::increment ()
{
  maybeFetchFirst();
  ROS_ASSERT_MSG(sub_ && sub_->next_result_, "Can't increment iterator past the end");
  sub_->fetchNext();
}

template <class M>
void
ResultIterator<M>::QueryResultSubscriber::fetchNext ()
{
  PullMessage srv;
  srv.request.query_id = query_id_;
  data_received_ = metadata_only_;
  metadata_received_ = false;
  boost::unique_lock<boost::mutex> l(mutex_);
  next_result_.reset(new MessageWithMetadata<M>());
  if (!pull_client_.call(srv))
    throw WarehouseClientException("Service call to pull_message failed");
  ROS_DEBUG_NAMED ("query_int", "pulling %u messages", srv.response.num_messages_sent);
  if (srv.response.num_messages_sent > 0) 
  {
    ROS_ASSERT(srv.response.num_messages_sent == 1);
    ROS_DEBUG_NAMED ("query_int", "Waiting for unique lock");
    while (!(data_received_ && metadata_received_) && ros::ok()) {
      ROS_DEBUG_NAMED ("query_int", "Waiting on condition variable");
      cond.wait(l);
    }
  }
  else
    next_result_.reset();
  ROS_DEBUG_NAMED ("query_int", "Done fetching");
}


template <class M>
void ResultIterator<M>::QueryResultSubscriber::messageCallback (typename M::ConstPtr msg)
{
  ROS_ASSERT_MSG (!data_received_, "Code assumption violated: data received though we didn't ask for it.");
  ROS_DEBUG_NAMED ("query_int", "Waiting for lock guard");
  boost::lock_guard<boost::mutex> l(mutex_);
  M* base = next_result_.get();
  *base = *msg;
  ROS_DEBUG_STREAM_NAMED ("query_int", "Received message " << *msg);
  data_received_ = true;
  if (metadata_received_)
    cond.notify_all();
}

template <class M>
void ResultIterator<M>::QueryResultSubscriber::metadataCallback (std_msgs::String::ConstPtr m)
{
  ROS_ASSERT_MSG (!metadata_received_, "Code assumption violated: data received though we didn't ask for it.");
  ROS_DEBUG_NAMED ("query_int", "Waiting for lock guard");
  boost::lock_guard<boost::mutex> l(mutex_);
  ROS_DEBUG_NAMED ("query_int", "Got lock guard");
  next_result_->metadata = m->data;
  ROS_DEBUG_NAMED ("query_int", "Received metadata %s", m->data.c_str());
  metadata_received_ = true;
  if (data_received_)
    cond.notify_all();
}


template <class M>
bool ResultIterator<M>::equal (const ResultIterator<M>& other) const
{
  /// Incomplete; the only case we care about is whether
  /// we're at the end yet.  
  if (sub_) {
    maybeFetchFirst();
    return !sub_->next_result_ && (!other.sub_ || !other.sub_->next_result_);
  }
  else
    return !other.sub_ || !other.sub_->next_result_;
}




} // namespace
