#!/usr/bin/env python
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import roslib; roslib.load_manifest('warehouse')
import rospy
import warehouse.srv as whs
from warehouse.msg import *
from warehouse.exceptions import *
import threading

class ClientCollectionHandler:
  """
  A handler for client connections to a collection in the warehouse.
  """

  def __init__(self, db_name, collection_name, msg, indexed_fields):
    """
    Constructor
    """
    self.db_name = db_name
    self.collection_name = collection_name

    self.pkg_type_name = msg._type[:msg._type.find("/")]
    self.msg_type_name = msg._type[ msg._type.find("/")+1:]

    self.push_condition = threading.Condition()
    self.query_condition = threading.Condition()

    self.notify_cb = None

    # Load the module that contains the specified message type
    try:
      roslib.load_manifest(self.pkg_type_name)
      m = __import__(self.pkg_type_name+".msg")
      mod = getattr(m,'msg')
    except roslib.packages.InvalidROSPkgException:
      rospy.logerr( 'package {0} not found'.format(self.pkg_type_name) )
    except ImportError:
      rospy.logerr( 'no messages found in {0}'.format(self.pkg_type_name) )

    # Look up message class
    try:
      self.msg_type = getattr(mod, self.msg_type_name)
    except AttributeError:
      rospy.logerr( 'message type {0} not found in {1}'.format(self.msg_type_name, self.pkg_type_name) )

    # Create a service to create this collection
    rospy.wait_for_service('setup_collection')
    create_srv = rospy.ServiceProxy('setup_collection', whs.SetupCollection)

    # Create a service to perform queries
    rospy.wait_for_service('initiate_query')
    self.query_srv = rospy.ServiceProxy('initiate_query', whs.InitiateQuery)

    # Create a service to pull messages from a query
    rospy.wait_for_service('pull_message')
    self.pull_srv = rospy.ServiceProxy('pull_message', whs.PullMessage)

    rospy.wait_for_service('close_collection')
    self.close_collection_srv = rospy.ServiceProxy('close_collection', whs.CloseCollection)

    # Create this collection
    resp = create_srv(db_name=self.db_name, collection_name=self.collection_name,
                      msg_pkg=self.pkg_type_name, msg_type=self.msg_type_name,
                      indexed_fields=indexed_fields)

    if resp.error_code == whs.SetupCollectionResponse.PACKAGE_NOT_FOUND:
      raise ClientException("Package not found: {0}".format(resp.error_msg))
    elif resp.error_code == whs.SetupCollectionResponse.MESSAGE_NOT_FOUND_IN_PACKAGE:
      raise ClientException("Message not found: {0}".format(resp.error_msg))
    elif resp.error_code == whs.SetupCollectionResponse.INVALID_INDEXED_FIELD:
      raise ClientException("Invalid indexed field: {0}".format(resp.error_msg))
    elif resp.error_code != whs.SetupCollectionResponse.SUCCESS:
      raise ClientException("Backend error: {0}".format(resp.error_msg))
    else:
        
      insert_topic_name = "warehouse/%s/%s/insert" % (self.db_name, self.collection_name)
      notify_topic_name =  "warehouse/%s/%s/notify" % (self.db_name, self.collection_name)
      # Create a publisher for the data insertion topic
      self.publisher = rospy.Publisher(insert_topic_name, self.msg_type, latch=True)
      # Create a subscriber on the the insertion notification topic
      self.subscriber = rospy.Subscriber(notify_topic_name, UpdateNotification, self._handle_notify)

  def close(self):
    if self.subscriber != None  and self.publisher != None :
      self.subscriber.unregister()
      self.close_collection_srv(db_name=self.db_name, collection_name=self.collection_name)
      self.publisher.unregister()
    self.subscriber = None
    self.publisher = None

  def push_data(self, msg):
    """
    Push data into a collection
    @param msg : The new data to insert into the database
    """
    # Send the data over the wire
    self.publisher.publish(msg)
    self.push_condition.acquire()

    # Wait for the insertion notification
    if self.push_condition.wait(10.0):
      rospy.logerr("Timeout reached while waiting for Warehouse response to a push data")
    self.push_condition.release()


  def subscribe_to_insertion(self, cb):
    """
    Register a function callback that is triggered whenever new data is inserted into the collection
    @param cb: The callback
    """
    self.notify_cb = cb
   

  def query(self, conditions):
    """
    Query a collection for data
    @param conditions : Conditions for the query
    @return a list of matching entries
    """
    result = []

    # Initialize the query
    init_resp = self.query_srv(db_name= self.db_name, collection_name=self.collection_name, query=conditions, metadata_only=False )

    # This is the topic on which results will be published
    topic = "warehouse/%s/%s/%s" % (self.db_name, self.collection_name, init_resp.query_id)
    query_sub = rospy.Subscriber(topic, self.msg_type, self._handle_pull, result)


    # Wait for the publisher to kick in
    while query_sub.get_num_connections() == 0:
      rospy.sleep(0.1)
   
    # Iterate through all the results
    done = False
    while not done:
      self.query_condition.acquire()
      pull_resp = self.pull_srv( db_name=self.db_name, collection_name=self.collection_name, query_id=init_resp.query_id )
      done = pull_resp.num_messages_sent == 0
      if not done:
        self.query_condition.wait(10.0)
        self.query_condition.release()

    query_sub.unregister()
    return result

  def _handle_notify(self, msg):
    """
    Handles data insertion notifications.
    """
    self.push_condition.acquire()
    rospy.loginfo("Data inserted into collection[{0}:{1}]".format(self.db_name, self.collection_name) )
    self.push_condition.notify()
    self.push_condition.release()
    if self.notify_cb != None:
      self.notify_cb( msg )
 

  def _handle_pull(self, msg, result):
    """
    Handle a pull from a query. This funciton is only used internally
    """
    self.query_condition.acquire()
    result.append(msg)
    self.query_condition.notify()
    self.query_condition.release()




#-------------------------------------------------------------------------------
# The Client API. Use this class to interface to the Warehouse.
class Client:
  """
  The warehouse library interface.
  """

  def __init__(self):
    """
    Initialize the warehouse library
    """
    self.collections = {}
    rospy.wait_for_service('list_collections')
    self.list_collections_srv = rospy.ServiceProxy('list_collections', whs.ListCollections)

    rospy.wait_for_service('drop_db')
    self.drop_db_srv = rospy.ServiceProxy('drop_db', whs.DropDB)

  def __del__(self):
    for k,v in self.collections.items():
      for c,h in v.items():
        h.close()

  def has_collection(self, db_name, collection_name):
    """
    Checks for the existence of a collection in a database.
    @param db_name : Name of a database
    @param collection_name : Name of a collection
    """
    resp = self.list_collections_srv(db_name=db_name)
    return (resp.db_exists, collection_name in resp.collections)

  def drop_db(self, db_name):
    """
    Drop a database. This will invalidate all connections to the database
    @param db_name : Name of the databse to delete
    """
    # Be nice and close all of our collections
    for k,v in self.collections[db_name].items():
      v.close()
      del self.collections[db_name][k]
    del self.collections[db_name]

    self.drop_db_srv(db_name=db_name)
    return True

  #def close_collection(self, collection):
  #  collection.close()
  #  d_name = collection.db_name
  #  c_name = collection.collection_name
  #  if d_name in self.collections and c_name in self.collections[d_name]:
  #    del self.collections[collection.db_name][collection.collection_name]

  #  if len(self.collections[d_name]) == 0:
  #    del self.collections[d_name]

  def setup_collection(self, db_name, collection_name, msg, indexed_fields):
    """
    Create a new collection in the specified database. The collection will contain documents of type <pkg_type>.<msg_type>
    @param db_name : Name of the database
    @param collection_name : Name of the collection
    @param msg : Message type
    @param indexed_fields: Fields on which to index the data. Each element must be a nested field specifier such as 'pose.position.x'
    @return A instance of connection to the specified collection
    """

    if not db_name in self.collections:
      self.collections[db_name] = {}

    if not collection_name in self.collections[db_name]:
      self.collections[db_name][collection_name] = ClientCollectionHandler(db_name, collection_name, msg, indexed_fields)

    return self.collections[db_name][collection_name]
