/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Implementation of WarehouseClient template methods
 *
 * Only meant to be included by warehouse_client.h
 *
 * \author Bhaskara Marthi
 */

#include <warehouse/SetupCollection.h>
#include <warehouse/exception.h>
#include <boost/algorithm/string.hpp>

namespace warehouse
{

namespace mt=ros::message_traits;
using std::string;
using std::vector;


template <class M>
Collection<M>
WarehouseClient::setupCollection (const string& collection,
                                  const StringVec& indexed,
                                  const StringVec& additional_metadata)
{
  return setupCollection<M> (collection, "", indexed,
      additional_metadata);
}

template <class M>
Collection<M>
WarehouseClient::setupCollection (const string& collection,
    const string& topic, const StringVec& indexed,
    const StringVec& additional_metadata)
{
  typedef mt::DataType<M> DataType;
  boost::mutex::scoped_lock l(mutex_);

  // Infer the message package and type
  const string data_type = DataType::value();
  vector<string> tokens;
  boost::split(tokens, data_type, boost::is_any_of(string("/")));
  ROS_ASSERT_MSG(tokens.size()==2, "Data type of message was %s, "
                 "which is invalid", data_type.c_str());

  // Construct service request
  SetupCollection srv;
  srv.request.db_name = db_;
  srv.request.collection_name = collection;
  srv.request.topic = topic;
  srv.request.msg_pkg = tokens[0];
  srv.request.msg_type = tokens[1];
  srv.request.indexed_fields = indexed;
  srv.request.additional_metadata_fields = additional_metadata;

  // Call service
  if (!setup_client_.call(srv)) 
    throw WarehouseClientException("Service call to setup_collection failed");    

  // Pass on errors
  if (srv.response.error_code != SetupCollection::Response::SUCCESS)
    throw SetupCollectionException(srv.response.error_msg);

  // Create Collection object
  ROS_DEBUG_STREAM_NAMED ("setup_collection", "Service call succeeded");
  const string actual_topic = topic.empty() ? 
    "warehouse/"+db_+"/"+collection+"/insert" :
    topic;

  return Collection<M>(actual_topic, db_, collection, nh_);
}



} // namespace
