/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Test script for warehouse client
 *
 * \author Bhaskara Marthi
 */

// %Tag(CPP_CLIENT)%

#include "client_test_helpers.h"
#include <warehouse/warehouse_client.h>
#include <yaml-cpp/yaml.h>
#include <ros/console.h>
#include <boost/foreach.hpp>
#include <gtest/gtest.h>

namespace wh=warehouse;
namespace gm=geometry_msgs;
using std::vector;
using std::string;

typedef wh::MessageWithMetadata<gm::Pose>::ConstPtr PoseWithMetadata;

string makeMetadata (const string& v)
{
  return wh::MetadataString().add("foo", v).add("bar", 42);
}


TEST(Warehouse, warehouseClient)
{ 
  wh::WarehouseClient client("my_app");

  /// Setup collection to store poses, and index them on
  /// position.x and position.y (they will also be
  /// automatically indexed on creation time)
  vector<std::string> indexed_fields, additional_metadata;
  indexed_fields.push_back("position.x");
  indexed_fields.push_back("position.y");
  
  wh::Collection<gm::Pose> coll;
  coll = client.setupCollection<gm::Pose>("poses", indexed_fields);

  /// Add some poses, together with some metadata for each pose
  const gm::Pose p1 = makePose(24, 42, 0);
  const gm::Pose p2 = makePose(10, 532, 3);
  const gm::Pose p3 = makePose(53, 22, 5);
  const gm::Pose p4 = makePose(22, -5, 33);
  coll.publish(p1, makeMetadata("bar"));
  coll.publish(p2, makeMetadata("baz"));
  coll.publish(p3, makeMetadata("qux"));
  coll.publish(p1, makeMetadata("oof"));
  coll.publish(p4, makeMetadata("ooof"));

  // Set up query: position.x < 40 and position.y > 0
  vector<wh::Condition> c(2);
  c[0].field_name = "position.x";
  c[0].predicate = wh::Condition::LESS_THAN;
  c[0].args.push_back("40");
  c[1].field_name = "position.y";
  c[1].predicate = wh::Condition::GREATER_THAN;
  c[1].args.push_back("0");

  // Arrange to get the results in reverse order of creation time
  wh::OrderingCriterion order;
  order.field = "_creation_time";
  order.reverse = true;

  // Execute query
  vector<PoseWithMetadata> poses = coll.pullAllResults(c, false, order);

  // Verify poses. 
  EXPECT_EQ(3u, poses.size());
  EXPECT_EQ(p1, *poses[0]);
  EXPECT_EQ(p2, *poses[1]);
  EXPECT_EQ(p1, *poses[2]);

  EXPECT_EQ("oof", wh::getMetadataVal<string>(poses[0]->metadata, "foo"));
  EXPECT_EQ("baz", wh::getMetadataVal<string>(poses[1]->metadata, "foo"));
  EXPECT_EQ("bar", wh::getMetadataVal<string>(poses[2]->metadata, "foo"));

  // Try creating another collection object with the same collection
  wh::Collection<gm::Pose> coll2 =
    client.setupCollection<gm::Pose>("poses", indexed_fields);

  // Delete some messages
  vector<wh::Condition> c2(1);
  c2[0].field_name = "position.y";
  c2[0].predicate = wh::Condition::LESS_THAN;
  c2[0].args.push_back("30");

  vector<wh::Condition> c3;
  EXPECT_EQ(5u, coll2.pullAllResults(c3, false).size());
  EXPECT_EQ(2u, coll2.removeMessages(c2));
  EXPECT_EQ(3u, coll2.pullAllResults(c3, true).size());

  coll.queryResults(c);
}


int main (int argc, char** argv)
{
  ros::init(argc, argv, "client_test");
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
 
// %EndTag(CPP_CLIENT)%
